"""Factor Analysis of Mixed Data (FAMD)"""
from __future__ import annotations

import numpy as np
import pandas as pd
import sklearn.utils
from sklearn import preprocessing

from prince import pca


class FAMD(pca.PCA):
    def __init__(
        self,
        n_components=2,
        n_iter=3,
        copy=True,
        check_input=True,
        random_state=None,
        engine="sklearn",
        handle_unknown="error",
    ):
        super().__init__(
            rescale_with_mean=True,
            rescale_with_std=False,
            n_components=n_components,
            n_iter=n_iter,
            copy=copy,
            check_input=check_input,
            random_state=random_state,
            engine=engine,
        )
        self.handle_unknown = handle_unknown

    def _check_input(self, X):
        if self.check_input:
            sklearn.utils.check_array(X, dtype=[str, np.number])

    def fit(self, X, y=None):
        # Separate numerical columns from categorical columns
        self.num_cols_ = X.select_dtypes(include=["float"]).columns.tolist()
        if not self.num_cols_:
            raise ValueError("All variables are qualitative: MCA should be used")
        self.cat_cols_ = X.columns.difference(self.num_cols_).tolist()
        if not self.cat_cols_:
            raise ValueError("All variables are quantitative: PCA should be used")

        # Preprocess numerical columns
        X_num = X[self.num_cols_].copy()
        self.num_scaler_ = preprocessing.StandardScaler().fit(X_num)
        X_num[:] = self.num_scaler_.transform(X_num)

        # Preprocess categorical columns
        X_cat = X[self.cat_cols_]
        self.cat_scaler_ = preprocessing.OneHotEncoder(handle_unknown=self.handle_unknown).fit(
            X_cat
        )
        X_cat_oh = pd.DataFrame.sparse.from_spmatrix(
            self.cat_scaler_.transform(X_cat),
            index=X_cat.index,
            columns=self.cat_scaler_.get_feature_names_out(self.cat_cols_),
        )
        prop = X_cat_oh.sum() / X_cat_oh.sum().sum() * 2
        X_cat_oh_norm = X_cat_oh.sub(X_cat_oh.mean(axis="rows")).div(prop**0.5, axis="columns")

        # PCA.fit doesn't work with sparse matrices. Well, it accepts them, but it densifies them.
        # We pre-densify them here to avoid a warning.
        # TODO: In the future, PCA should be able to handle sparse matrices.
        X_cat_oh_norm = X_cat_oh_norm.sparse.to_dense()

        Z = pd.concat([X_num, X_cat_oh_norm], axis=1)
        super().fit(Z)

        # Determine column_coordinates_
        # This is based on line 184 in FactoMineR's famd.R file
        rc = self.row_coordinates(X)
        weights = np.ones(len(X_cat_oh)) / len(X_cat_oh)
        norm = (rc**2).multiply(weights, axis=0).sum()
        eta2 = pd.DataFrame(index=rc.columns)
        for i, col in enumerate(self.cat_cols_):
            # TODO: there must be a better way to select a subset of the one-hot encoded matrix
            tt = X_cat_oh[[f"{col}_{i}" for i in self.cat_scaler_.categories_[i]]]
            ni = (tt / len(tt)).sum()
            eta2[col] = (
                rc.apply(lambda x: (tt.multiply(x * weights, axis=0).sum() ** 2 / ni).sum()) / norm
            ).values
        self.column_coordinates_ = pd.concat(
            [self.column_coordinates_.loc[self.num_cols_] ** 2, eta2.T]
        )
        self.column_coordinates_.columns.name = "component"
        self.column_coordinates_.index.name = "variable"

        return self

    def row_coordinates(self, X):
        # Separate numerical columns from categorical columns
        X_num = X[self.num_cols_].copy()
        X_cat = X[self.cat_cols_]

        # Preprocess numerical columns
        X_num[:] = self.num_scaler_.transform(X_num)

        # Preprocess categorical columns
        X_cat = pd.DataFrame.sparse.from_spmatrix(
            self.cat_scaler_.transform(X_cat),
            index=X_cat.index,
            columns=self.cat_scaler_.get_feature_names_out(self.cat_cols_),
        )
        prop = X_cat.sum() / X_cat.sum().sum() * 2
        X_cat = X_cat.sub(X_cat.mean(axis="rows")).div(prop**0.5, axis="columns")

        Z = pd.concat([X_num, X_cat], axis=1)

        return super().row_coordinates(Z)

    def inverse_transform(self, X):
        raise NotImplementedError("FAMD inherits from PCA, but this method is not implemented yet")

    def row_standard_coordinates(self, X):
        raise NotImplementedError("FAMD inherits from PCA, but this method is not implemented yet")

    def row_cosine_similarities(self, X):
        raise NotImplementedError("FAMD inherits from PCA, but this method is not implemented yet")

    def column_correlations(self, X):
        raise NotImplementedError("FAMD inherits from PCA, but this method is not implemented yet")

    def column_cosine_similarities_(self, X):
        raise NotImplementedError("FAMD inherits from PCA, but this method is not implemented yet")

    @property
    def column_contributions_(self):
        return self.column_coordinates_ / self.eigenvalues_



from __future__ import annotations

import pathlib

import pandas as pd

DATASETS_DIR = pathlib.Path(__file__).parent / "datasets"


def load_energy_mix(year=2019, normalize=True):
    """Per capita energy mix by country in 2019.

    Each row corresponds to a country. There is one column for each energy source.
    A value corresponds to the average energy consumption of a source per capita.
    For instance, in France, every citizen consumed 15,186 kWh of nuclear energy.

    This data comes from https://ourworldindata.org/energy-mix

    Parameters
    ----------
    year
        The year the study was made.
    normalize
        Whether or not to normalize the kWh by country.

    """

    df = (
        pd.read_csv(DATASETS_DIR / "per-capita-energy-stacked.csv")
        .query("Year == @year")
        .query("Entity not in ['Africa', 'Europe', 'North America', 'World']")
        .drop(columns=["Code", "Year"])
        .rename(columns={"Entity": "Country"})
        .rename(columns=lambda x: x.replace(" per capita (kWh)", "").lower())
        .set_index(["continent", "country"])
    )
    if normalize:
        return df.div(df.sum(axis="columns"), axis="rows")
    return df


def load_decathlon():
    """The Decathlon dataset from FactoMineR."""
    decathlon = pd.read_csv(DATASETS_DIR / "decathlon.csv")
    decathlon.columns = ["athlete", *map(str.lower, decathlon.columns[1:])]
    decathlon.athlete = decathlon.athlete.apply(str.title)
    decathlon = decathlon.set_index(["competition", "athlete"])
    return decathlon


def load_french_elections():
    """Voting data for the 2022 French elections, by region.

    The [original dataset](https://www.data.gouv.fr/fr/datasets/resultats-du-premier-tour-de-lelection-presidentielle-2022-par-commune-et-par-departement/#resources)
    has been transformed into a contingency matrix. The latter tallies the number of votes for the
    12 candidates across all 18 regions. The number of blank and abstentions are also recorded.
    More information about these regions, including a map, can be found
    [on Wikipedia](https://www.wikiwand.com/fr/Région_française).

    """
    dataset = pd.read_csv(DATASETS_DIR / "02-resultats-par-region.csv")
    cont = dataset.pivot(index="reg_name", columns="cand_nom", values="cand_nb_voix")
    cont["Abstention"] = dataset.groupby("reg_name")["abstention_nb"].min()
    cont["Blank"] = dataset.groupby("reg_name")["blancs_nb"].min()
    cont.columns = [c.title() for c in cont.columns]
    cont.index.name = "region"
    cont.columns.name = "candidate"
    return cont


def load_punctuation_marks():
    """Punctuation marks of six French writers."""
    return pd.read_csv(DATASETS_DIR / "punctuation_marks.csv", index_col="author")


def load_hearthstone_cards():
    """Hearthstone standard cards.

    Source: https://gist.github.com/MaxHalford/32ed2c80672d7391ec5b4e6f291f14c1

    """
    return pd.read_csv(DATASETS_DIR / "hearthstone_cards.csv", index_col="id")


def load_burgundy_wines():
    """Burgundy wines dataset.

    Source: https://personal.utdallas.edu/~herve/Abdi-MCA2007-pretty.pdf

    """
    wines = pd.DataFrame(
        data=[
            [1, 6, 7, 2, 5, 7, 6, 3, 6, 7],
            [5, 3, 2, 4, 4, 4, 2, 4, 4, 3],
            [6, 1, 1, 5, 2, 1, 1, 7, 1, 1],
            [7, 1, 2, 7, 2, 1, 2, 2, 2, 2],
            [2, 5, 4, 3, 5, 6, 5, 2, 6, 6],
            [3, 4, 4, 3, 5, 4, 5, 1, 7, 5],
        ],
        columns=pd.MultiIndex.from_tuples(
            [
                ("Expert 1", "Fruity"),
                ("Expert 1", "Woody"),
                ("Expert 1", "Coffee"),
                ("Expert 2", "Red fruit"),
                ("Expert 2", "Roasted"),
                ("Expert 2", "Vanillin"),
                ("Expert 2", "Woody"),
                ("Expert 3", "Fruity"),
                ("Expert 3", "Butter"),
                ("Expert 3", "Woody"),
            ],
            names=("expert", "aspect"),
        ),
        index=[f"Wine {i + 1}" for i in range(6)],
    )
    wines.insert(0, "Oak type", [1, 2, 2, 2, 1, 1])
    return wines


def load_beers():
    """Beers dataset.

    The data is taken from https://github.com/philipperemy/beer-dataset.

    """
    return pd.read_csv(DATASETS_DIR / "beers.csv.zip", index_col="name")



"""Multiple Factor Analysis (MFA)"""
from __future__ import annotations

import collections

import altair as alt
import numpy as np
import pandas as pd
import sklearn.utils

from prince import pca, utils


class MFA(pca.PCA, collections.UserDict):
    def __init__(
        self,
        n_components=2,
        n_iter=3,
        copy=True,
        check_input=True,
        random_state=None,
        engine="sklearn",
    ):
        super().__init__(
            rescale_with_mean=False,
            rescale_with_std=False,
            n_components=n_components,
            n_iter=n_iter,
            copy=copy,
            check_input=check_input,
            random_state=random_state,
            engine=engine,
        )
        collections.UserDict.__init__(self)

    def _check_input(self, X):
        if self.check_input:
            sklearn.utils.check_array(X, dtype=[str, np.number])

    def fit(self, X, y=None, groups=None):

        # Checks groups are provided
        self.groups_ = self._determine_groups(X, groups)

        # Check input
        self._check_input(X)

        # Check group types are consistent
        self.all_nums_ = {}
        for name, cols in sorted(self.groups_.items()):
            all_num = all(pd.api.types.is_numeric_dtype(X[c]) for c in cols)
            all_cat = all(pd.api.types.is_string_dtype(X[c]) for c in cols)
            if not (all_num or all_cat):
                raise ValueError(f'Not all columns in "{name}" group are of the same type')
            self.all_nums_[name] = all_num

        # Normalize column-wise
        X = (X - X.mean()) / ((X - X.mean()) ** 2).sum() ** 0.5

        # Run a factor analysis in each group
        for name, cols in sorted(self.groups_.items()):
            if self.all_nums_[name]:
                fa = pca.PCA(
                    rescale_with_mean=True,
                    rescale_with_std=True,
                    n_components=self.n_components,
                    n_iter=self.n_iter,
                    copy=True,
                    random_state=self.random_state,
                    engine=self.engine,
                )
            else:
                raise NotImplementedError("Groups of non-numerical variables are not supported yet")
            self[name] = fa.fit(X.loc[:, cols])

        # Fit the global PCA
        Z = pd.concat(
            (X[cols] / self[g].eigenvalues_[0] ** 0.5 for g, cols in self.groups_.items()),
            axis="columns",
        )
        super().fit(Z)
        self.total_inertia_ = sum(self.eigenvalues_)

        # TODO: column_coordinates_ is not implemented yet
        delattr(self, "column_coordinates_")

        return self

    def _determine_groups(self, X, provided_groups):
        if provided_groups is None:
            raise ValueError("Groups have to be specified")
        if isinstance(provided_groups, list):
            if not isinstance(X.columns, pd.MultiIndex):
                raise ValueError("Groups have to be provided as a dict when X is not a MultiIndex")
            groups = {
                g: [
                    (g, c)
                    for c in X.columns.get_level_values(1)[X.columns.get_level_values(0) == g]
                ]
                for g in provided_groups
            }
        else:
            groups = provided_groups
        return groups

    @property
    @utils.check_is_fitted
    def eigenvalues_(self):
        """Returns the eigenvalues associated with each principal component."""
        return np.square(self.svd_.s)

    def row_coordinates(self, X):
        """Returns the row principal coordinates."""

        if (X.index != self.row_contributions_.index).any():
            raise NotImplementedError("Supplementary rows are not supported yet")

        X = (X - X.mean()) / ((X - X.mean()) ** 2).sum() ** 0.5
        Z = pd.concat(
            (X[cols] / self[g].eigenvalues_[0] ** 0.5 for g, cols in self.groups_.items()),
            axis="columns",
        )
        U = self.svd_.U
        s = self.svd_.s
        M = np.full(len(X), 1 / len(X))

        return (Z @ Z.T) @ (M[:, np.newaxis] ** (-0.5) * U * s**-1)

    def group_row_coordinates(self, X):

        if (X.index != self.row_contributions_.index).any():
            raise NotImplementedError("Supplementary rows are not supported yet")

        X = (X - X.mean()) / ((X - X.mean()) ** 2).sum() ** 0.5
        Z = pd.concat(
            (X[cols] / self[g].eigenvalues_[0] ** 0.5 for g, cols in self.groups_.items()),
            axis="columns",
        )
        M = np.full(len(X), 1 / len(X))
        U = self.svd_.U
        s = self.svd_.s

        def add_index(g, group_name):
            g.columns = pd.MultiIndex.from_tuples(
                [(group_name, col) for col in g.columns],
                names=("group", "component"),
            )
            return g

        return len(self.groups_) * pd.concat(
            [
                add_index(
                    g=(Z[g] @ Z[g].T) @ (M[:, np.newaxis] ** (-0.5) * U * s**-1),
                    group_name=g,
                )
                for g, cols in self.groups_.items()
            ],
            axis="columns",
        )

    def column_coordinates(self, X):
        raise NotImplementedError("MFA inherits from PCA, but this method is not implemented yet")

    def inverse_transform(self, X):
        raise NotImplementedError("MFA inherits from PCA, but this method is not implemented yet")

    def row_standard_coordinates(self, X):
        raise NotImplementedError("MFA inherits from PCA, but this method is not implemented yet")

    def row_cosine_similarities(self, X):
        raise NotImplementedError("MFA inherits from PCA, but this method is not implemented yet")

    def column_correlations(self, X):
        raise NotImplementedError("MFA inherits from PCA, but this method is not implemented yet")

    def column_cosine_similarities_(self, X):
        raise NotImplementedError("MFA inherits from PCA, but this method is not implemented yet")

    @property
    def column_contributions_(self):
        raise NotImplementedError("MFA inherits from PCA, but this method is not implemented yet")

    def plot(self, X, x_component=0, y_component=1, color_by=None, **params):

        if color_by is not None:
            params["color"] = color_by

        params["tooltip"] = (X.index.names if isinstance(X.index, pd.MultiIndex) else ["index"]) + [
            f"component {x_component}",
            f"component {y_component}",
        ]

        eig = self._eigenvalues_summary.to_dict(orient="index")

        row_coords = self.row_coordinates(X)
        row_coords.columns = [f"component {i}" for i in row_coords.columns]
        row_coords = row_coords.reset_index()
        row_plot = (
            alt.Chart(row_coords)
            .mark_circle()
            .encode(
                alt.X(
                    f"component {x_component}",
                    scale=alt.Scale(zero=False),
                    axis=alt.Axis(
                        title=f"component {x_component} — {eig[x_component]['% of variance'] / 100:.2%}"
                    ),
                ),
                alt.Y(
                    f"component {y_component}",
                    scale=alt.Scale(zero=False),
                    axis=alt.Axis(
                        title=f"component {y_component} — {eig[y_component]['% of variance'] / 100:.2%}"
                    ),
                ),
                **params,
            )
        )

        return row_plot.interactive()



from __future__ import annotations

import functools

import altair as alt
import numpy as np
import pandas as pd
from sklearn.utils import validation


def check_is_fitted(method):
    """Decorator"""

    @functools.wraps(method)
    def _impl(self, *method_args, **method_kwargs):
        validation.check_is_fitted(self)
        return method(self, *method_args, **method_kwargs)

    return _impl


def make_labels_and_names(X):

    if isinstance(X, pd.DataFrame):
        row_label = X.index.name if X.index.name else "Rows"
        row_names = X.index.tolist()
        col_label = X.columns.name if X.columns.name else "Columns"
        col_names = X.columns.tolist()
    else:
        row_label = "Rows"
        row_names = list(range(X.shape[0]))
        col_label = "Columns"
        col_names = list(range(X.shape[1]))

    return row_label, row_names, col_label, col_names


class EigenvaluesMixin:
    @property
    @check_is_fitted
    def percentage_of_variance_(self):
        """Returns the percentage of explained inertia per principal component."""
        return 100 * self.eigenvalues_ / self.total_inertia_

    @property
    @check_is_fitted
    def cumulative_percentage_of_variance_(self):
        """Returns the percentage of explained inertia per principal component."""
        return np.cumsum(self.percentage_of_variance_)

    @property
    @check_is_fitted
    def _eigenvalues_summary(self):
        """Return a summary of the eigenvalues and their importance."""
        return pd.DataFrame(
            {
                "eigenvalue": self.eigenvalues_,
                r"% of variance": self.percentage_of_variance_,
                r"% of variance (cumulative)": self.cumulative_percentage_of_variance_,
            },
            index=pd.RangeIndex(0, len(self.eigenvalues_), name="component"),
        )

    @property
    def eigenvalues_summary(self):
        """Return a summary of the eigenvalues and their importance."""
        summary = self._eigenvalues_summary
        summary["% of variance"] /= 100
        summary["% of variance (cumulative)"] /= 100
        summary["eigenvalue"] = summary["eigenvalue"].map("{:,.3f}".format)
        summary["% of variance"] = summary["% of variance"].map("{:.2%}".format)
        summary["% of variance (cumulative)"] = summary["% of variance (cumulative)"].map(
            "{:.2%}".format
        )
        summary.index.name = "component"
        return summary

    def scree_plot(self):
        """Scree plot.

        References
        ----------
        https://en.wikipedia.org/wiki/Scree_plot

        """
        eig = self._eigenvalues_summary.reset_index()
        eig["component"] = eig["component"].astype(str)
        return (
            alt.Chart(
                self._eigenvalues_summary.reset_index().assign(
                    component=lambda x: x["component"].astype(str)
                )
            )
            .mark_bar(size=10)
            .encode(x="component", y="eigenvalue", tooltip=eig.columns.tolist())
        )



"""Multiple Correspondence Analysis (MCA)"""
from __future__ import annotations

import numpy as np
import pandas as pd
import sklearn.base
import sklearn.utils

from prince import utils

from . import ca


class MCA(sklearn.base.BaseEstimator, sklearn.base.TransformerMixin, ca.CA):
    def fit(self, X, y=None):
        """Fit the MCA for the dataframe X.

        The MCA is computed on the indicator matrix (i.e. `X.get_dummies()`). If some of the columns are already
        in indicator matrix format, you'll want to pass in `K` as the number of "real" variables that it represents.
        (That's used for correcting the inertia linked to each dimension.)
        """

        if self.check_input:
            sklearn.utils.check_array(X, dtype=[str, np.number])

        # K is the number of actual variables, to apply the Benzécri correction
        self.K_ = X.shape[1]

        # One-hot encode the data
        one_hot = pd.get_dummies(X, columns=X.columns)

        # We need the number of columns to apply the Greenacre correction
        self.J_ = one_hot.shape[1]

        # Apply CA to the indicator matrix
        super().fit(one_hot)

        return self

    def row_coordinates(self, X):
        return super().row_coordinates(pd.get_dummies(X, columns=X.columns))

    def row_cosine_similarities(self, X):
        oh = pd.get_dummies(X, columns=X.columns)
        return super()._row_cosine_similarities(X=oh, F=super().row_coordinates(oh))

    def column_coordinates(self, X):
        return super().column_coordinates(pd.get_dummies(X, columns=X.columns))

    def column_cosine_similarities(self, X):
        oh = pd.get_dummies(X, columns=X.columns)
        return super()._column_cosine_similarities(X=oh, G=super().column_coordinates(oh))

    @utils.check_is_fitted
    def transform(self, X):
        """Computes the row principal coordinates of a dataset."""
        if self.check_input:
            sklearn.utils.check_array(X, dtype=[str, np.number])
        return self.row_coordinates(X)



"""Correspondence Analysis (CA)"""
from __future__ import annotations

import functools

import altair as alt
import numpy as np
import pandas as pd
from scipy import sparse
from sklearn.utils import check_array

from prince import svd, utils


def select_active_columns(method):
    @functools.wraps(method)
    def _impl(self, X=None, *method_args, **method_kwargs):
        if hasattr(self, "active_cols_") and isinstance(X, pd.DataFrame):
            return method(self, X[self.active_cols_], *method_args, **method_kwargs)
        return method(self, X, *method_args, **method_kwargs)

    return _impl


def select_active_rows(method):
    @functools.wraps(method)
    def _impl(self, X=None, *method_args, **method_kwargs):
        if hasattr(self, "active_rows_") and isinstance(X, pd.DataFrame):
            return method(self, X.loc[self.active_rows_], *method_args, **method_kwargs)
        return method(self, X, *method_args, **method_kwargs)

    return _impl


class CA(utils.EigenvaluesMixin):
    def __init__(
        self,
        n_components=2,
        n_iter=10,
        copy=True,
        check_input=True,
        random_state=None,
        engine="sklearn",
    ):
        self.n_components = n_components
        self.n_iter = n_iter
        self.copy = copy
        self.check_input = check_input
        self.random_state = random_state
        self.engine = engine

    def fit(self, X, y=None):
        # Check input
        if self.check_input:
            check_array(X)

        # Check all values are positive
        if (X < 0).any().any():
            raise ValueError("All values in X should be positive")

        _, row_names, _, col_names = utils.make_labels_and_names(X)

        if isinstance(X, pd.DataFrame):
            X = X.to_numpy()

        if self.copy:
            X = np.copy(X)

        # Compute the correspondence matrix which contains the relative frequencies
        X = X.astype(float) / np.sum(X)

        # Compute row and column masses
        self.row_masses_ = pd.Series(X.sum(axis=1), index=row_names)
        self.col_masses_ = pd.Series(X.sum(axis=0), index=col_names)

        self.active_rows_ = self.row_masses_.index.unique()
        self.active_cols_ = self.col_masses_.index.unique()

        # Compute standardised residuals
        r = self.row_masses_.to_numpy()
        c = self.col_masses_.to_numpy()
        S = sparse.diags(r**-0.5) @ (X - np.outer(r, c)) @ sparse.diags(c**-0.5)

        # Compute SVD on the standardised residuals
        self.svd_ = svd.compute_svd(
            X=S,
            n_components=min(self.n_components, min(X.shape) - 1),
            n_iter=self.n_iter,
            random_state=self.random_state,
            engine=self.engine,
        )

        # Compute total inertia
        self.total_inertia_ = np.einsum("ij,ji->", S, S.T)

        self.row_contributions_ = pd.DataFrame(
            sparse.diags(self.row_masses_.values)
            @ (
                # Same as row_coordinates(X)
                (
                    sparse.diags(self.row_masses_.values**-0.5)
                    @ self.svd_.U
                    @ sparse.diags(self.svd_.s)
                )
                ** 2
            )
            / self.eigenvalues_,
            index=self.row_masses_.index,
        )

        self.column_contributions_ = pd.DataFrame(
            sparse.diags(self.col_masses_.values)
            @ (
                # Same as col_coordinates(X)
                (
                    sparse.diags(self.col_masses_.values**-0.5)
                    @ self.svd_.V.T
                    @ sparse.diags(self.svd_.s)
                )
                ** 2
            )
            / self.eigenvalues_,
            index=self.col_masses_.index,
        )

        return self

    @property
    @utils.check_is_fitted
    def eigenvalues_(self):
        """Returns the eigenvalues associated with each principal component."""
        return np.square(self.svd_.s)

    @select_active_columns
    def row_coordinates(self, X):
        """The row principal coordinates."""

        _, row_names, _, _ = utils.make_labels_and_names(X)
        index_name = X.index.name

        if isinstance(X, pd.DataFrame):
            try:
                X = X.sparse.to_coo().astype(float)
            except AttributeError:
                X = X.to_numpy()

        if self.copy:
            X = X.copy()

        # Normalise the rows so that they sum up to 1
        if isinstance(X, np.ndarray):
            X = X / X.sum(axis=1)[:, None]
        else:
            X = X / X.sum(axis=1)

        return pd.DataFrame(
            data=X @ sparse.diags(self.col_masses_.to_numpy() ** -0.5) @ self.svd_.V.T,
            index=pd.Index(row_names, name=index_name),
        )

    @select_active_columns
    def row_cosine_similarities(self, X):
        """Return the cos2 for each row against the dimensions.

        The cos2 value gives an indicator of the accuracy of the row projection on the dimension.

        Values above 0.5 usually means that the row is relatively accurately well projected onto that dimension. Its often
        used to identify which factor/dimension is important for a given element as the cos2 can be interpreted as the proportion
        of the variance of the element attributed to a particular factor.

        """
        F = self.row_coordinates(X)
        return self._row_cosine_similarities(X, F)

    @select_active_columns
    def _row_cosine_similarities(self, X, F):
        # Active
        X_act = X.loc[self.active_rows_]
        X_act = X_act / X_act.sum().sum()
        marge_col = X_act.sum(axis=0)
        Tc = X_act.div(X_act.sum(axis=1), axis=0).div(marge_col, axis=1) - 1
        dist2_row = (Tc**2).mul(marge_col, axis=1).sum(axis=1)

        # Supplementary
        X_sup = X.loc[X.index.difference(self.active_rows_, sort=False)]
        X_sup = X_sup.div(X_sup.sum(axis=1), axis=0)
        dist2_row_sup = ((X_sup - marge_col) ** 2).div(marge_col, axis=1).sum(axis=1)

        dist2_row = pd.concat((dist2_row, dist2_row_sup))

        # Can't use pandas.div method because it doesn't support duplicate indices
        return F**2 / dist2_row.to_numpy()[:, None]

    @select_active_rows
    def column_coordinates(self, X):
        """The column principal coordinates."""

        _, _, _, col_names = utils.make_labels_and_names(X)
        index_name = X.columns.name

        if isinstance(X, pd.DataFrame):
            is_sparse = X.dtypes.apply(pd.api.types.is_sparse).all()
            if is_sparse:
                X = X.sparse.to_coo()
            else:
                X = X.to_numpy()

        if self.copy:
            X = X.copy()

        # Transpose and make sure the rows sum up to 1
        if isinstance(X, np.ndarray):
            X = X.T / X.T.sum(axis=1)[:, None]
        else:
            X = X.T / X.T.sum(axis=1)

        return pd.DataFrame(
            data=X @ sparse.diags(self.row_masses_.to_numpy() ** -0.5) @ self.svd_.U,
            index=pd.Index(col_names, name=index_name),
        )

    @select_active_rows
    def column_cosine_similarities(self, X):
        """Return the cos2 for each column against the dimensions.

        The cos2 value gives an indicator of the accuracy of the column projection on the dimension.

        Values above 0.5 usually means that the column is relatively accurately well projected onto that dimension. Its often
        used to identify which factor/dimension is important for a given element as the cos2 can be interpreted as the proportion
        of the variance of the element attributed to a particular factor.
        """
        G = self.column_coordinates(X)
        return self._column_cosine_similarities(X, G)

    @select_active_rows
    def _column_cosine_similarities(self, X, G):
        # Active
        X_act = X[self.active_cols_]
        X_act = X_act / X_act.sum().sum()
        marge_row = X_act.sum(axis=1)
        Tc = X_act.div(marge_row, axis=0).div(X_act.sum(axis=0), axis=1) - 1
        dist2_col = (Tc**2).mul(marge_row, axis=0).sum(axis=0)

        # Supplementary
        X_sup = X[X.columns.difference(self.active_cols_, sort=False)]
        X_sup = X_sup.div(X_sup.sum(axis=0), axis=1)
        dist2_col_sup = ((X_sup.sub(marge_row, axis=0)) ** 2).div(marge_row, axis=0).sum(axis=0)

        dist2_col = pd.concat((dist2_col, dist2_col_sup))
        return (G**2).div(dist2_col, axis=0)

    @utils.check_is_fitted
    def plot(self, X, x_component=0, y_component=1, **params):
        row_coords = self.row_coordinates(X)
        row_coords.columns = [f"component {i}" for i in row_coords.columns]
        row_coords = row_coords.assign(
            variable=row_coords.index.name or "row", value=row_coords.index.astype(str)
        )

        col_coords = self.column_coordinates(X)
        col_coords.columns = [f"component {i}" for i in col_coords.columns]
        col_coords = col_coords.assign(
            variable=col_coords.index.name or "column",
            value=col_coords.index.astype(str),
        )

        coords = pd.concat([row_coords, col_coords])
        eig = self._eigenvalues_summary.to_dict(orient="index")

        return (
            alt.Chart(coords)
            .mark_circle(size=50)
            .encode(
                alt.X(
                    f"component {x_component}",
                    scale=alt.Scale(zero=False),
                    axis=alt.Axis(
                        title=f"component {x_component} — {eig[x_component]['% of variance'] / 100:.2%}"
                    ),
                ),
                alt.Y(
                    f"component {y_component}",
                    scale=alt.Scale(zero=False),
                    axis=alt.Axis(
                        title=f"component {y_component} — {eig[y_component]['% of variance'] / 100:.2%}"
                    ),
                ),
                color="variable",
                tooltip=[
                    "variable",
                    "value",
                    f"component {x_component}",
                    f"component {y_component}",
                ],
                **params,
            )
            .interactive()
        )



"""Generalized Procrustes Analysis (GPA)"""
from __future__ import annotations

import numpy as np
from scipy.linalg import orthogonal_procrustes
from scipy.spatial import procrustes
from sklearn import base, utils


class GPA(base.BaseEstimator, base.TransformerMixin):
    """Generalized Procrustes Analysis (GPA).

    Algorithm outline:
        (https://wikipedia.org/wiki/Generalized_Procrustes_analysis)
    1. Choose a reference shape.
    2. Apply Procrustes Analysis to superimpose all shapes to the reference
        shape.
    3. Compute the mean shape of the superimposed shapes.
    4. Repeat steps 2 and 3 until convergence.

    Parameters:
        max_iter (int): The maximum number of Procrustes analysis iterations.
        tol (float): The tolerance for the optimization; stops if the
            procrustes distance decreases by less or equal to ``tol`` between
            iterations.
        init ({'random', 'mean'}): Method for initializing reference shape.
            'random' : choose reference shape from shape list
            'mean' : initialize reference shape as mean of shape list
        scale (bool): Whether to compute transformations with a scale component
        copy (bool): Whether to copy data or perform the computations inplace.
            If False, data passed to fit are overwritten and running
            fit(X).transform(X) will not yield the expected results,
            use fit_transform(X) instead.
        check_input (bool): Whether to check the consistency of the inputs.
        random_state (int, RandomState instance or None): Determines random
            number generation for initialization when ``init=='random'``

    """

    def __init__(
        self,
        max_iter=10,
        tol=1e-4,
        init="random",
        scale=True,
        copy=True,
        check_input=True,
        random_state=None,
    ):
        self.max_iter = max_iter
        self.tol = tol
        self.init = init
        self.scale = scale
        self.copy = copy
        self.check_input = check_input
        self.random_state = random_state

    def fit(self, X, y=None):
        """Fit the model with X.

        The algorithm naturally fits and transforms at the same time, so this
        simply calls ``.fit_transform``

        Parameters:
            X (array-like of shape (n_shapes, n_points, n_dim)): Matrix of
                shapes to match to each other.
            y: Ignored

        Returns:
            self (object): The instance itself
        """
        self.fit_transform(X)

        return self

    def transform(self, X):
        """Align X to the reference shape.

        Parameters:
            X (array-like of shape (n_shapes, n_points, n_dim)): Matrix of
                shapes to align to the refernce shape.

        Returns:
            X_new (array-like of shape (n_shapes, n_points, n_dim)): Matrix of
                aligned shapes
        """
        self._check_is_fitted()
        if self.check_input:
            self._check_input(X)

        X_new = np.empty(X.shape)
        for shape_idx in range(X.shape[0]):
            _, X_new[shape_idx], _ = procrustes(self.reference_shape, X[shape_idx])

        return X_new

    def fit_transform(self, X, y=None):
        """Fit the model with X and return the aligned shapes.

        Parameters:
            X (array-like of shape (n_shapes, n_points, n_dim)): Matrix of
                shapes to match to each other.
            y: Ignored

        Returns:
            X_new (array-like of shape (n_shapes, n_points, n_dim)): Matrix X
                of aligned shapes
        """

        # Check input
        if self.check_input:
            self._check_input(X)

        # Copy data
        if self.copy:
            X = np.array(X, copy=True)

        # scikit-learn SLEP010
        n_shapes, n_points, n_dim = X.shape
        self.n_features_in_ = n_dim

        # Pick reference shape
        if self.init == "random":
            random_state = utils.check_random_state(self.random_state)
            ref_shape_idx = random_state.randint(X.shape[0])
            reference_shape = X[ref_shape_idx].copy()
        elif self.init == "mean":
            reference_shape = X.mean(axis=0)
        else:
            raise ValueError("init method must be one of ('random', 'mean')")

        for iter_idx in range(self.max_iter):
            # Align each shape to reference shape
            for shape_idx in range(X.shape[0]):
                if self.scale:
                    _, X[shape_idx], _ = procrustes(reference_shape, X[shape_idx])
                else:
                    _, X[shape_idx] = unscaled_procrustes(reference_shape, X[shape_idx])

            # Compute diagnostics
            mean_shape = X.mean(axis=0)
            procrustes_distance = np.linalg.norm(reference_shape - mean_shape)

            # Update reference shape
            reference_shape = mean_shape

            # Check for convergence
            if procrustes_distance <= self.tol:
                break

        # Store properties
        self._reference_shape = reference_shape

        # Return the aligned shapes
        return X

    def _check_input(self, X):
        utils.check_array(X, allow_nd=True)
        if X.ndim != 3:
            raise ValueError("Expected 3-dimensional input of (n_shapes, n_points, n_dim)")

    def _check_is_fitted(self):
        utils.validation.check_is_fitted(self, "_reference_shape")

    @property
    def reference_shape(self):
        """Returns the final reference shape."""
        self._check_is_fitted()
        return self._reference_shape


def unscaled_procrustes(reference, data):
    """Fit `data` to `reference` using procrustes analysis without scaling.
    Uses translation (mean-centering), reflection, and orthogonal rotation.

    Parameters:
        reference (array-like of shape (n_points, n_dim)): reference shape to
            fit `data` to
        data (array-like of shape (n_points, n_dim)): shape to align to
            `reference`

    Returns:
        reference_centered (np.ndarray of shape (n_points, n_dim)): 0-centered
            `reference` shape
        data_aligned (np.ndarray of shape (n_points, n_dim)): `data` aligned to
            the reference shape
    """
    # Convert inputs to np.ndarray types
    reference = np.array(reference, dtype=np.double)
    data = np.array(data, dtype=np.double)

    # Translate data to the origin
    reference_centered = reference - reference.mean(axis=0)
    data_centered = data - data.mean(axis=0)

    # Rotate / reflect data to match reference
    # transform mtx2 to minimize disparity
    R, _ = orthogonal_procrustes(data_centered, reference_centered)
    data_aligned = data_centered @ R

    return reference_centered, data_aligned
